#include <errno.h>
#include <rrd.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define LOG_ERR(S, ...) do { if (errno != 0) \
	fprintf(stderr, "error: " S " (%s)\n", ##__VA_ARGS__, strerror(errno)); \
	else \
	fprintf(stderr, "error: " S "\n", ##__VA_ARGS__); \
} while(0);

static void print_usage() {
	printf("Usage: rrdsplit [-d] [-o outfile1.rrd[,outfile2.rrd[,...]] source.rrd\n"
		"-h prints this stuff\n"
		"-o specifies a comma separated list of output RRD filenames, an empty element instructs rrdsplit to ignore the corresponding datasource\n"
		"-d will remove source.rrd if the conversion went well\n"
		"\n"
		"Examples:\n"
		"rrdsplit -o outfile1,outfile2.rrd source.rrd         - splits DS1 and DS2 from source.rrd into outfile1.rrd and outfile2.rrd, respectively\n"
		"rrdsplit -o outfile1.rrd,,outfile3.rrd source.rrd    - splits DS1 and DS3 from source.rrd into outfile1.rrd and outfile3.rrd, respectively, ignoring DS2\n"
		"rrdsplit -o ,outfile2.rrd source.rrd                 - splits DS2 from source.rrd into outfile2.rrd, ignoring DS1\n"
		"rrdsplit -d -o outfile1,outfile2.rrd source.rrd      - splits DS1 and DS2 from source.rrd into outfile1.rrd and outfile2.rrd, respectively and unlinks source.rrd\n"
		"rrdsplit source.rrd                                  - splits each datasource from source.rrd into an rrd named split_ds-<ds_idx>.rrd where <ds_idx> is the index of the datasource in source.rrd\n"
		);
}
/* These aren't declared in any distributed header file */
#define RRD_READONLY    (1<<0)
#define RRD_READWRITE   (1<<1)
#define RRD_CREAT       (1<<2)
#define RRD_READAHEAD   (1<<3)
#define RRD_COPY        (1<<4)
#define RRD_EXCL        (1<<5)
#define RRD_READVALUES  (1<<6)

static bool write_all(rrd_file_t *f, const void *buf, size_t count)
{
	ssize_t written = 0;
	do {
		ssize_t n = 0;
		if ((n = rrd_write(f, buf + written, count - written)) < 0) {
			if (errno == EAGAIN)
				continue;
			else {
				LOG_ERR("failed to write to RRD file");
				return false;
			}
		}

		written += n;

	} while (written < count);

	return true;
}

static int split_one(rrd_t rrd, rrd_file_t *src_rrd_file, const char *rrd_out_filename, unsigned long ds_idx) {
	rrd_t rrdnew;
	rrd_file_t *rrd_out_file;
	rrd_init(&rrdnew);
	/* These need to be initialised before calling rrd_open() with
	   the RRD_CREATE flag */

	if ((rrdnew.stat_head = (stat_head_t*)calloc(1, sizeof(stat_head_t))) == NULL) {
		LOG_ERR("allocating stat_head for new RRD");
		return (-1);
	}

	/*
	 * Use the source RRD as a template for our new single-DS RRD,
	 * but make sure to update it to reflect the cardinality
	 */
	memcpy(rrdnew.stat_head, rrd.stat_head, sizeof(stat_head_t));
	rrdnew.stat_head->ds_cnt = 1;


	/* The RRA definitions should be the same */
	if ((rrdnew.rra_def = malloc(sizeof(rra_def_t) * rrd.stat_head->rra_cnt)) == NULL) {
		LOG_ERR("could not allocate rra_def for new RRD");
		rrd_free(&rrdnew);
		return (-1);
	}
	memcpy(rrdnew.rra_def, rrd.rra_def,sizeof(rra_def_t) * rrd.stat_head->rra_cnt);

	rrd_out_file = rrd_open(rrd_out_filename, &rrdnew, RRD_READWRITE | RRD_CREAT);
	if (rrd_out_file == NULL) {
		LOG_ERR("could not create '%s': %s", rrd_out_filename, strerror(errno));
		rrd_free(&rrdnew);
		return (-1);
	}

	if (rrd_lock(rrd_out_file) != 0) {
		LOG_ERR("could not lock new RRD %s: %s", rrd_out_filename, strerror(errno));
		goto err_out;
	}

	if ((rrdnew.rra_ptr = malloc(sizeof(rra_ptr_t) * rrd.stat_head->rra_cnt)) == NULL) {
		LOG_ERR("could not allocate rra_ptr for new RRD");
		rrd_free(&rrdnew);
		rrd_close(rrd_out_file);
		return (-1);
	}

	rrdnew.ds_def = &rrd.ds_def[ds_idx];
	rrdnew.ds_def[0].ds_nam[0] = '1';
	rrdnew.ds_def[0].ds_nam[1] = '\0';
	rrdnew.live_head = rrd.live_head;
	rrdnew.pdp_prep = &rrd.pdp_prep[ds_idx];
	rrdnew.cdp_prep = &rrd.cdp_prep[ds_idx];
	memcpy(rrdnew.rra_ptr, rrd.rra_ptr,
			sizeof(rra_ptr_t) * rrd.stat_head->rra_cnt
		  );

#define WRITE_OR_DIE(File, Buf, Cnt) if (!write_all(File, Buf, Cnt)) goto err_out;

	WRITE_OR_DIE(rrd_out_file, rrdnew.stat_head, sizeof(stat_head_t));
	WRITE_OR_DIE(rrd_out_file, rrdnew.ds_def, sizeof(ds_def_t));
	WRITE_OR_DIE(rrd_out_file, rrdnew.rra_def, sizeof(rra_def_t) * rrdnew.stat_head->rra_cnt);
	WRITE_OR_DIE(rrd_out_file, rrdnew.live_head, sizeof(live_head_t) * 1);
	WRITE_OR_DIE(rrd_out_file, rrdnew.pdp_prep, sizeof(pdp_prep_t));

	for (size_t rra_idx = 0; rra_idx < rrd.stat_head->rra_cnt; ++rra_idx) {
		WRITE_OR_DIE(rrd_out_file, rrdnew.cdp_prep + (rra_idx * rrd.stat_head->ds_cnt), sizeof(cdp_prep_t));
	}
	WRITE_OR_DIE(rrd_out_file, rrdnew.rra_ptr, sizeof(rra_ptr_t) * rrdnew.stat_head->rra_cnt);

	off_t rra_start = src_rrd_file->header_len;
	rrd_seek(src_rrd_file, rra_start, SEEK_SET);

	/* XXX: This could be made more efficient, surely. But is it worth the effort? */
	for (size_t rra_idx = 0; rra_idx < rrd.stat_head->rra_cnt; ++rra_idx) { /* For each RRA ... */
		for (size_t ix = 0; ix < rrd.rra_def[rra_idx].row_cnt; ++ix) { /* And each row ... */
			rrd_value_t buffer;
			rrd_seek(src_rrd_file, ds_idx * sizeof(rrd_value_t), SEEK_CUR); /*skip to next value for this DS in this row*/
			rrd_read(src_rrd_file, &buffer, sizeof(rrd_value_t)); /* read it */
			WRITE_OR_DIE(rrd_out_file, &buffer, sizeof(rrd_value_t)); /* write it */
			rrd_seek(src_rrd_file, (rrd.stat_head->ds_cnt - 1 - ds_idx) * sizeof(rrd_value_t), SEEK_CUR); /*... and skip past the rest of the row*/
		}
	}

	rrd_close(rrd_out_file);
	printf("%s\n", rrd_out_filename);
	rrdnew.ds_def = NULL;
	rrdnew.live_head = NULL;
	rrdnew.pdp_prep = NULL;
	rrdnew.cdp_prep = NULL;
	rrd_free(&rrdnew);
	return 0;
err_out:
	rrd_close(rrd_out_file);
	rrdnew.ds_def = NULL;
	rrdnew.live_head = NULL;
	rrdnew.pdp_prep = NULL;
	rrdnew.cdp_prep = NULL;

	rrd_free(&rrdnew);
	return -1;
}
static int split_rrd(rrd_t rrd, rrd_file_t *src_rrd_file, const char **outfilenames, size_t count)
{
    for (size_t i = 0; i < count; i++) {
	if (outfilenames[i] != NULL) {
	    if (split_one(rrd, src_rrd_file, outfilenames[i], i) < 0)
		return -1;
	}
    }
    return 0;
}

char **parse_outfiles(const char *outfile_str, unsigned long
	*outfiles_count) {
    const char *p, *last_p;
    char **outfiles = NULL;
    char *new_outfile = NULL;
    unsigned long i;

    p = last_p = outfile_str;

    *outfiles_count = 0;
    while ((p = strchrnul(p, ','))) {
	outfiles = realloc(outfiles, (*outfiles_count + 1) * sizeof(*outfiles));
	if (p == outfile_str || *(p-1) == ',') {
	    outfiles[*outfiles_count] = NULL;
	} else {
	    new_outfile = strndup(last_p, p-last_p);
	    outfiles[*outfiles_count] = new_outfile;
	    if (new_outfile == NULL) {
		LOG_ERR("failed to allocate memory for output filename");
		goto err_out;
	    }

	    for (i = 0; i < *outfiles_count; ++i) {
		if (outfiles[i] != NULL && (strcmp(new_outfile, outfiles[i]) == 0)) {
		    LOG_ERR("duplicate definition of output filename %s", new_outfile);
		    goto err_out;
		}
	    }
	}
	++*outfiles_count;
	if (*p == '\0')
	    break;

	++p;
	last_p = p;
    }
    return outfiles;
err_out:
    for (i = 0; i <= *outfiles_count; ++i) {
	free(outfiles[i]);
    }
    free(outfiles);
    return NULL;

}

int main(int argc, char **argv)
{
	int opt;
	char **outfiles = NULL;
	char *sourcefile = NULL;
	unsigned long ds_out_cnt = 0;
	bool delete_rrd_source = false;
	while ((opt = getopt(argc, argv, "dho:")) != -1) {
		switch (opt) {
		    case 'h':
			print_usage();
			return EXIT_SUCCESS;
		    case 'd':
			delete_rrd_source = true;
			break;
		    case 'o':
			if (outfiles != NULL) {
			    LOG_ERR("option -o can only be specified once, ignoring second instance");
			    break;
			}
			if ((outfiles = parse_outfiles(optarg, &ds_out_cnt)) == NULL) {
			    return EXIT_FAILURE;
			}
			break;
		    default:
			free(outfiles);
			print_usage();
			return EXIT_FAILURE;
		}
	}

	if (optind >= argc) {
		print_usage();
		free(outfiles);
		return EXIT_FAILURE;
	}

	sourcefile = strdup(argv[optind]);
	struct rrd_t rrd;
	rrd_init(&rrd);

	rrd_file_t *src_rrd_file = rrd_open(sourcefile,
		&rrd,
		RRD_READWRITE
		);

	if (src_rrd_file == NULL) {
		LOG_ERR("could not open source RRD");
		free(outfiles);
		rrd_free(&rrd);
		free(sourcefile);
		return EXIT_FAILURE;
	}

	if (ds_out_cnt == 0 && outfiles == NULL) { /* No filenames specified*/
		outfiles = calloc(rrd.stat_head->ds_cnt, sizeof(char *));
		for (size_t i = 0; i < rrd.stat_head->ds_cnt; ++i) {
			if (asprintf(&outfiles[i], "split_ds-%lu.rrd", i+1) < 0) {
				LOG_ERR("could not allocate rrd filename (DS %zu)", i);
				rrd_free(&rrd);
				for (i; i > 0; --i) {
				    free(outfiles[i-1]);
				}
				free(outfiles);
				rrd_close(src_rrd_file);
				free(sourcefile);
				return EXIT_FAILURE;
			}
		}
		ds_out_cnt = rrd.stat_head->ds_cnt;
	}

	if (rrd_lock(src_rrd_file) != 0) {
		LOG_ERR("could not lock source RRD");
		for (size_t i = 0; i < ds_out_cnt; ++i) {
		    free(outfiles[i]);
		}
		free(outfiles);
		rrd_free(&rrd);
		rrd_close(src_rrd_file);
		free(sourcefile);
		return EXIT_FAILURE;
	}


	int ret = split_rrd(rrd, src_rrd_file, (const char **)outfiles, ds_out_cnt > rrd.stat_head->ds_cnt ? rrd.stat_head->ds_cnt : ds_out_cnt);
	rrd_free(&rrd);
	rrd_close(src_rrd_file);

	for (size_t i = 0; i < ds_out_cnt; ++i) {
		free(outfiles[i]);
	}

	free(outfiles);

	if (0 > ret) {
		LOG_ERR("failed to split RRD");
		free(sourcefile);
		return EXIT_FAILURE;
	}

	if(delete_rrd_source && unlink(sourcefile) == -1) {
		// not important enough to change exit code
		LOG_ERR("failed to delete source file even if it was requested");
	}
	free(sourcefile);

	return EXIT_SUCCESS;
}
